@@ -1,7 +1,7 @@
 /* ------------------------------------------------------------------------- */
 
 /*
- * Copyright 2007-2015 GRAHAM DUMPLETON
+ * Copyright 2007-2016 GRAHAM DUMPLETON
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,6 +21,7 @@
 #include "wsgi_logger.h"
 
 #include "wsgi_server.h"
+#include "wsgi_metrics.h"
 
 /* ------------------------------------------------------------------------- */
 
@@ -449,6 +450,14 @@ static PyObject *Log_writable(LogObject *self, PyObject *args)
     Py_INCREF(Py_True);
     return Py_True;
 }
+
+static PyObject *Log_fileno(LogObject *self, PyObject *args)
+{
+    PyErr_SetString(PyExc_IOError, "Apache/mod_wsgi log object is not "
+            "associated with a file descriptor.");
+
+    return NULL;
+}
 #endif
 
 static PyObject *Log_closed(LogObject *self, void *closure)
@@ -504,6 +513,7 @@ static PyMethodDef Log_methods[] = {
     { "readable",   (PyCFunction)Log_readable,   METH_NOARGS, 0 },
     { "seekable",   (PyCFunction)Log_seekable,   METH_NOARGS, 0 },
     { "writable",   (PyCFunction)Log_writable,   METH_NOARGS, 0 },
+    { "fileno",     (PyCFunction)Log_fileno,   METH_NOARGS, 0 },
 #endif
     { NULL, NULL}
 };
@@ -563,7 +573,8 @@ PyTypeObject Log_Type = {
     0,                      /*tp_is_gc*/
 };
 
-void wsgi_log_python_error(request_rec *r, PyObject *log, const char *filename)
+void wsgi_log_python_error(request_rec *r, PyObject *log,
+                           const char *filename, int publish)
 {
     PyObject *m = NULL;
     PyObject *result = NULL;
@@ -672,9 +683,26 @@ void wsgi_log_python_error(request_rec *r, PyObject *log, const char *filename)
         }
     }
     else {
-        Py_XDECREF(type);
-        Py_XDECREF(value);
-        Py_XDECREF(traceback);
+        if (publish) {
+            PyObject *event = NULL;
+            PyObject *object = NULL;
+
+            if (wsgi_event_subscribers()) {
+                event = PyDict_New();
+
+                object = Py_BuildValue("(OOO)", type, value, traceback);
+                PyDict_SetItemString(event, "exception_info", object);
+                Py_DECREF(object);
+
+                wsgi_publish_event("request_exception", event);
+
+                Py_DECREF(event);
+            }
+        }
+
+        Py_DECREF(type);
+        Py_DECREF(value);
+        Py_DECREF(traceback);
     }
 
     Py_XDECREF(result);
